# -*- coding: utf8 -*-
from __future__ import unicode_literals

import csv
import datetime
import json
import logging

import six
from six import BytesIO, StringIO

from driverng import MessageV2, Worklist
from driverng.splitted import Splitted, String

logger = logging.getLogger("driverng.chronos_ananke")


class ChronosAnankeBase(Splitted):

    SEPARATOR = ","


class ChronosAnankeResult(object):
    FIELDS = [
        String("chamber_number"),
        String("chamber_sample_name"),
        String("chamber_sample_type"),
        String("channel_1_fluor"),
        String("channel_1_target_name"),
        String("channel_1_type"),
        String("channel_1_cq"),
        String("channel_2_fluor"),
        String("channel_2_target_name"),
        String("channel_2_type"),
        String("channel_2_cq"),
        String("channel_3_fluor"),
        String("channel_3_target_name"),
        String("channel_3_type"),
        String("channel_3_cq"),
        String("channel_4_fluor"),
        String("channel_4_target_name"),
        String("channel_4_type"),
        String("channel_4_cq"),
        String("interpretation"),
    ]

    def parse(self, line):
        fields = self.FIELDS
        for index, field in enumerate(fields):
            setattr(self, field.name, line[index])


class ChronosAnankeMultiple(list):
    _MESSAGE_TYPE_MAP = {MessageV2.RESULT: ChronosAnankeResult}

    def add(self, type_message=None, **kwargs):
        obj = self._MESSAGE_TYPE_MAP[type_message](**kwargs)
        self.append(obj)
        return obj

    def parse(self, data):
        if six.PY2:
            buf = BytesIO(data.encode("utf-8"))
        else:
            buf = StringIO(data)
        reader = csv.reader(buf, delimiter=str(","))
        self.operator_user = None
        self.experiment_name = None
        self.note = None
        self.test_type = None
        self.reagent_id = None
        self.chip_id = None
        self.start_date = None
        self.chronos_instrument = None
        for row in reader:
            if six.PY2:
                line = [c.decode("utf-8") for c in row]
            else:
                line = row
            if not len(line):
                continue
            if line[0] == "Export created by user":
                self.operator_user = line[1].strip()
            if line[0] == "Experiment name":
                self.experiment_name = line[1].strip()
            if line[0] == "Note":
                self.note = line[1].strip()
            if line[0] == "Barcodes":
                barcodes = json.loads(
                    "{}, {}, {}".format(
                        line[1].strip(), line[2].strip(), line[3].strip()
                    )
                )
                self.reagent_id = barcodes.get("reagent_id")
                self.chip_id = barcodes.get("chip_id")
            if line[0] == "Test type":
                self.test_type = line[1].strip()
            if line[0] == "Start date":
                self.start_date = datetime.datetime.strptime(
                    line[1].strip(), "%Y-%m-%d %H:%M:%S"
                )
            if line[0] == "Chronos instrument":
                self.chronos_instrument = line[1].strip()
            if len(line) >= 6:
                obj = self.add(type_message=MessageV2.RESULT)
                obj.parse(line)


class Driver(MessageV2):
    SPEC = {"code": "chronos_ananke", "label": "Chronos DX BforCure Ananke"}

    OPTIONS = [
        MessageV2.Option("analyseur_traites", label="Liste des analyseurs traités")
    ]

    def __init__(self, **kwargs):
        analyseur_traites = kwargs.pop("analyseur_traites", None)
        self.analyseur_traites = (
            analyseur_traites.split(",") if analyseur_traites else None
        )
        super(Driver, self).__init__(**kwargs)

    def parse(self, data):
        obj = ChronosAnankeMultiple()
        obj.parse(data)
        return obj

    def decode_message(self, message):
        worklist = Worklist(context="result")
        for m in message:
            if (
                self.analyseur_traites is not None
                and message.chronos_instrument not in self.analyseur_traites
            ):
                logger.error(
                    "L'instrument %s n'est pas traité par la connexion"
                    % message.chronos_instrument
                )
                continue
            lots = []
            if message.reagent_id and message.chip_id:
                lots.append(
                    {
                        "type": "reagent",
                        "lot": "{}_{}".format(message.chip_id, message.reagent_id),
                    }
                )
            if m.chamber_number == "Chamber number":
                continue
            if m.chamber_sample_name != "":
                sample = worklist.get_or_create(
                    sample_id=m.chamber_sample_name,
                    rack=m.chamber_number,
                    type="patient",
                )
            if (
                m.chamber_sample_name != ""
                and m.channel_1_fluor != "N/A"
                and m.channel_1_target_name != ""
            ):
                sample.results.add(
                    code="{}_{}_1".format(m.channel_1_fluor, m.channel_1_target_name),
                    value=m.channel_1_cq,
                    instrument=message.chronos_instrument,
                    date=message.start_date,
                    operator=message.operator_user,
                    comments=[message.note],
                    lots=lots,
                    flags=[m.interpretation],
                )
            if (
                m.chamber_sample_name != ""
                and m.channel_2_fluor != "N/A"
                and m.channel_2_target_name != ""
            ):
                sample.results.add(
                    code="{}_{}_2".format(m.channel_2_fluor, m.channel_2_target_name),
                    value=m.channel_2_cq,
                    instrument=message.chronos_instrument,
                    date=message.start_date,
                    operator=message.operator_user,
                    comments=[message.note],
                    lots=lots,
                    flags=[m.interpretation],
                )
            if (
                m.chamber_sample_name != ""
                and m.channel_3_fluor != "N/A"
                and m.channel_3_target_name != ""
            ):
                sample.results.add(
                    code="{}_{}_3".format(m.channel_3_fluor, m.channel_3_target_name),
                    value=m.channel_3_cq,
                    instrument=message.chronos_instrument,
                    date=message.start_date,
                    operator=message.operator_user,
                    comments=[message.note],
                    lots=lots,
                    flags=[m.interpretation],
                )
            if (
                m.chamber_sample_name != ""
                and m.channel_4_fluor != "N/A"
                and m.channel_4_target_name != ""
            ):
                sample.results.add(
                    code="{}_{}_4".format(m.channel_4_fluor, m.channel_4_target_name),
                    value=m.channel_4_cq,
                    instrument=message.chronos_instrument,
                    date=message.start_date,
                    operator=message.operator_user,
                    comments=[message.note],
                    lots=lots,
                    flags=[m.interpretation],
                )
            if message.operator_user:
                sample.results.get_or_create(
                    code="Export created by user", value=message.operator_user
                )
            if message.experiment_name:
                sample.results.get_or_create(
                    code="Experiment name", value=message.experiment_name
                )
            if message.note:
                sample.results.get_or_create(code="Note", value=message.note)
            if message.test_type:
                sample.results.get_or_create(code="Test type", value=message.test_type)
        return worklist
