# -*- coding: utf8 -*-
from __future__ import unicode_literals

import base64
import logging
import time

from driverng import Worklist
from driverng.astm import AstmMessageV2, HeaderSegment
from driverng.astm import HprimDate as HL7Date
from driverng.astm import HprimForeign as HL7Foreign
from driverng.astm import HprimList
from driverng.astm import HprimMultiple
from driverng.astm import HprimString as HL7String
from driverng.astm import Segment

logger = logging.getLogger("driverng.hl7_clarilab_clinique")


class HL7ClarilabCliniqueMSH(HeaderSegment):
    sending_application = HL7String(3)
    sending_facility = HL7String(4)
    receiving_application = HL7String(5)
    receiving_facility = HL7String(6)
    date = HL7Date(7)
    # OML^O21: Admission
    # ORL^O22: ACK
    # ORU^R01: Result
    message_type = HprimMultiple(
        9,
        message_code=HL7String(1),
        trigger_event=HL7String(2),
        message_structure=HL7String(3),
    )
    control_id = HL7String(10)
    processing_id = HL7String(11, default="P")
    version_id = HL7String(12, default="2.5.1")
    country_code = HL7String(17, default="FRA")
    character_set = HL7String(18, default="UTF-8")

    patients = HL7Foreign("PID", "header")


class HL7ClarilabCliniquePID(Segment):
    patient_id = HL7String((4, 1))
    assigning_authority = (4, 4)
    identifier_type_code = HL7String((4, 5), default="PI")
    assigning_facility = HL7String((4, 6))
    last_name = HL7String((6, 1))
    first_name = HL7String((6, 2))
    further_names = HL7String((6, 3))
    prefix = HL7String((6, 5))
    type_code = HL7String((6, 7), default="L")
    maiden_name = HL7String(7)
    birthdate = HL7Date(8)
    sex = HL7String(9)
    adress_line_1 = HL7String((12, 1))
    adress_line_2 = HL7String((12, 2))
    city = HL7String((12, 3))
    postal_code = HL7String((12, 5))
    phone_numer = HprimList(14)
    home_phone_number = HL7String((14, 12))
    account_number = HL7String(19)
    identity_reliability_code = HL7String(33)

    visits = HL7Foreign("PV1", "patient")


class HL7ClarilabCliniquePV1(Segment):
    # N = Not applicable
    # I = Inpatient (hospitalisation)
    patient_class = HL7String(3, default="I")
    location = HL7String((4, 1))
    room = HL7String((4, 2))
    bed = HL7String((4, 3))
    facility = HL7String((4, 4))
    location_status = HL7String((4, 5))
    attending_id = HL7String((8, 1))
    attending_name = HL7String((8, 2))
    attending_firstname = HL7String((8, 3))
    attending_prefix = HL7String((8, 6))
    attending_type_code = HL7String((8, 10))
    visit_number = HL7String((20, 1))
    assigning_authority = HL7String((20, 4))
    identifier_type_code = HL7String((20, 5))
    admit_date = HL7Date(45)

    common_order = HL7Foreign("ORC", "patient")
    visits = HL7Foreign("PV2", "patient")


class HL7ClarilabCliniquePV2(Segment):
    visit_user_code = HL7String(8)
    expected_discharge_date = HL7Date(10)

    common_order = HL7Foreign("ORC", "patient")


class HL7ClarilabCliniqueORC(Segment):
    # NW = New order
    # SC = Status Change (résultats)
    order_control = HL7String(2)
    placer_group_number = HL7String(5)
    # CM = Order is completed
    order_status = HL7String(6, default="CM")
    transaction_date = HL7Date(10)
    ordering_providing_id = HL7String((13, 1))
    ordering_providing_name = HL7String((13, 2))
    ordering_providing_firstname = HL7String((13, 3))
    ordering_providing_prefix = HL7String((13, 6))
    ordering_providing_coding = HL7String((13, 10))
    codification = HL7String((13, 10))
    prescription_date = HL7Date(38)
    demand_identifier = HL7String(39)

    orders = HL7Foreign("OBR", "patient")
    sample_priority = HL7Foreign("TQ1", "patient")

    def to_string(self, separators=None):
        # Order control
        self.rang_segment = self.order_control
        return super(HL7ClarilabCliniqueORC, self).to_string(separators=separators)


class HL7ClarilabCliniqueOBR(Segment):
    place_order_number = HL7String(3)
    test_id = HL7String((5, 1))
    test_name = HL7String((5, 2))
    coding_name = HL7String((5, 3), default="LN")
    date = HL7Date(9)
    collector_id = HL7String((11, 1))
    collector_last_name = HL7String((11, 2))
    collector_first_name = HL7String((11, 3))
    collector_prefix = HL7String((11, 6))
    collector_id_coding = HL7String((11, 10))
    prescriber_id = HL7String((17, 1))
    prescriber_last_name = HL7String((17, 2))
    prescriber_first_name = HL7String((17, 3))
    prescriber_prefix = HL7String((17, 6))
    prescriber_type_code = HL7String((17, 10))
    result_copies_to_id = HL7String((29, 1))
    result_copies_to_name = HL7String((29, 2))
    result_copies_to_given_name = HL7String((29, 3))
    result_copies_to_authority = HL7String((29, 9))
    observer = HL7String(33)

    results = HL7Foreign("OBX", "order")
    sample_priority = HL7Foreign("TQ1", "order")
    specimens = HL7Foreign("SPM", "order")


class HL7ClarilabCliniqueOBX(Segment):
    MAX_VALUE_SUBFIELD = 5
    value_type = HL7String(3)
    observation_identifier = HL7String((4, 1))
    observation_text = HL7String((4, 2))
    coding_system = HL7String((4, 3), default="LN")
    alternate_identifier = HL7String((4, 4))
    alternate_text = HL7String((4, 5))
    alternate_coding_system = HL7String((4, 6))
    value = HprimMultiple(
        6,
        **dict(
            ("line{}".format(i), HL7String(i)) for i in range(1, MAX_VALUE_SUBFIELD + 1)
        )
    )
    units = HL7String(7)
    result_status = HL7String(12)
    date = HL7Date(15)
    validator = HL7String(17)
    instrument = HL7String(19)


class HL7ClarilabCliniqueSPM(Segment):
    specimen_id = HL7String(3)
    specimen_type = HL7String((5, 1))
    specimen_label = HL7String((5, 2))
    specimen_table = HL7String((5, 3))
    additive = HL7String(7)
    specimen_role = HL7String(12)
    collection_date = HL7Date(18)
    received_date = HL7Date(19)
    availability = HL7String(21)
    number_of_tubes = HL7String(27)

    infos_samples = HL7Foreign("SAC", "patient")


class HL7ClarilabCliniqueTQ1(Segment):
    priority = HL7String(10)

    orders = HL7Foreign("OBR", "patient")


class HL7ClarilabCliniqueSAC(Segment):
    carrier_type = HL7String(10)
    carrier_identifier = HL7String(11)
    cap_type = HL7String(27)


class HL7ClarilabCliniqueMSA(Segment):
    code = HL7String(2)
    control_id = HL7String(3)

    def to_string(self, separators=None):
        # code
        self.rang_segment = self.code
        return super(HL7ClarilabCliniqueMSA, self).to_string(separators=separators)


class Driver(AstmMessageV2):
    SPEC = {"code": "hl7_clarilab_clinique", "label": "HL7 PaLM CLARILAB CLINIQUE"}
    ENCODING = "utf-8"

    SEGMENT_CLASS_MAP = {
        "MSH": HL7ClarilabCliniqueMSH,
        "PID": HL7ClarilabCliniquePID,
        "PV1": HL7ClarilabCliniquePV1,
        "PV2": HL7ClarilabCliniquePV2,
        "ORC": HL7ClarilabCliniqueORC,
        "OBR": HL7ClarilabCliniqueOBR,
        "OBX": HL7ClarilabCliniqueOBX,
        "SPM": HL7ClarilabCliniqueSPM,
        "TQ1": HL7ClarilabCliniqueTQ1,
        "SAC": HL7ClarilabCliniqueSAC,
        "MSA": HL7ClarilabCliniqueMSA,
    }
    DEFAULT_SEPARATORS = "|^~\\&"
    FIRST_SEP_INDEX = 3
    SEGMENT_ENDING = "\r"

    OPTIONS = [
        AstmMessageV2.Option(
            "hl7_clarilab_clinique_sender_code", label="Code émetteur", default=""
        ),
        AstmMessageV2.Option(
            "hl7_clarilab_clinique_receiver_code", label="Code récepteur", default=""
        ),
    ]

    def __init__(self, **kwargs):
        self.sender_code = kwargs.pop("hl7_clarilab_clinique_sender_code", "")
        self.receiver_code = kwargs.pop("hl7_clarilab_clinique_receiver_code", "")
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        return HL7ClarilabCliniqueMSH(type_segment="MSH", **kwargs)

    def to_string(self, message, separators=None, **kwargs):
        return super(Driver, self).to_string(message, separators, **kwargs) + "\r"

    def encode_worklist(self, worklist, context):
        message = self.create(
            sending_application=self.sender_code,
            receiving_application=self.receiver_code,
            date=context["now"],
            control_id=context.get("control_id", int(time.time() * 1000)),
        )
        message.message_type.message_code = "ORU"
        message.message_type.trigger_event = "R01"
        message.message_type.message_structure = "ORU_R01"
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            priority={"S": "S", "A": "S", worklist.ELSE: "R"},
            action={"new": "NW"},
        ):
            seg_pid = message.patients.add(
                patient_id=patient.patient_id,
                last_name=patient.last_name,
                first_name=patient.first_name,
                prefix=patient.title,
                maiden_name=patient.maiden_name,
                birthdate=patient.birthdate,
                sex=patient.sex,
                city=patient.city,
                postal_code=patient.postal_code,
                account_number=patient.alternate_patient_id,
            )
            if patient.address:
                if "~" in patient.address:
                    seg_pid.adress_line_1, seg_pid.adress_line_2 = patient.address.split(
                        "~", 1
                    )
                else:
                    seg_pid.adress_line_1 = patient.address
            if patient.phone:
                if patient.phone == "T":
                    # Spécifique Clarilab
                    seg_pid.set_raw("telephone", "~T")
                else:
                    seg_pid.phone_numer = [p for p in patient.phone.split(";") if p]
            prescriber = [p for p in patient.physicians if p.type == "prescriber"]
            seg_pv1 = seg_pid.visits.add(
                location=patient.location, visit_number=patient.alternate_patient_id
            )
            seg_orc = seg_pv1.common_order.add(
                transaction_date=context["now"], order_control="SC"
            )
            for sample, action, priority in patient.samples:
                for result in sample.results:
                    seg_obr = seg_orc.orders.add(
                        place_order_number=sample.request_id,
                        test_id=result.code,
                        test_name=result.label,
                        date=result.date,
                        prescriber_id=prescriber[0].code if prescriber else "",
                        prescriber_last_name=prescriber[0].last_name
                        if prescriber
                        else "",
                        observer=self.sender_code,
                    )
                    unit = ""
                    if result.unit:
                        unit = result.unit
                    seg_obx = seg_obr.results.add(
                        value_type=result.type,
                        observation_identifier=result.code,
                        observation_text=result.label,
                        units=unit,
                        result_status="F",  # final
                        date=result.date,
                        validator=result.validator,
                        instrument=result.instrument,
                    )
                    values = result.value
                    if isinstance(values, (list, tuple)):
                        for i, value in enumerate(values, 1):
                            if i > seg_obx.MAX_VALUE_SUBFIELD:
                                raise RuntimeError(
                                    "Cannot encode %s: too much fields" % (values,)
                                )
                            setattr(seg_obx.value, "line%d" % (i,), value)
                    else:
                        seg_obx.value.line1 = values

        return message

    def decode_message(self, message):
        if (message.message_type.message_code, message.message_type.trigger_event) != (
            "OML",
            "O21",
        ):
            raise NotImplementedError

        worklist = Worklist(
            context="order",
            sender_id=message.sending_application,
            receiver_id=message.receiving_application,
        )

        for seg_pid in message.patients:
            for seg_pv1 in seg_pid.visits:
                if seg_pv1.visits:
                    for seg_pv2 in seg_pv1.visits:
                        common_orders = seg_pv2.common_order
                else:
                    common_orders = seg_pv1.common_order
                for seg_orc in common_orders:
                    sample = worklist.get_or_create(
                        type="patient",
                        sample_id=seg_orc.placer_group_number,
                        request_id=seg_orc.placer_group_number,
                    )
                    if seg_orc.order_control == "NW":
                        sample.request_date = seg_orc.transaction_date

                    further_names = []
                    if seg_pid.further_names:
                        for name in seg_pid.further_names.split(","):
                            further_names.append(name)
                    ins_number = ""
                    oid = ""
                    if seg_pid.identity_reliability_code == "VALI":
                        ins_number = seg_pid.patient_id
                        oid = seg_pid.assigning_facility
                    sample.update(
                        patient_id=seg_pid.patient_id,
                        ins_number=ins_number,
                        oid=oid,
                        identity_status=seg_pid.identity_reliability_code,
                        alternate_patient_id=seg_pv1.visit_number,
                        last_name=seg_pid.last_name,
                        first_name=seg_pid.first_name,
                        further_names=further_names,
                        birthdate=seg_pid.birthdate,
                        maiden_name=seg_pid.maiden_name,
                        title=seg_pid.prefix,
                        sex=seg_pid.sex,
                        address="{}~{}".format(
                            seg_pid.adress_line_1, seg_pid.adress_line_2
                        ),
                        postal_code=seg_pid.postal_code,
                        city=seg_pid.city,
                        phone=seg_pid.home_phone_number,
                        location=seg_pv1.location,
                        room=seg_pv1.room,
                        bed=seg_pv1.bed,
                        admission_date=seg_pv1.admit_date,
                    )
                    if seg_orc.ordering_providing_name:
                        sample.physicians.get_or_create(
                            type="prescriber",
                            code=seg_orc.ordering_providing_id,
                            last_name=seg_orc.ordering_providing_name,
                            firt_name=seg_orc.ordering_providing_firstname,
                            title=seg_orc.ordering_providing_prefix,
                            codification=seg_orc.ordering_providing_coding,
                        )
                    if seg_pv1.attending_name:
                        sample.physicians.get_or_create(
                            type="attending",
                            code=seg_pv1.attending_id,
                            last_name=seg_pv1.attending_name,
                            firt_name=seg_pv1.attending_firstname,
                            title=seg_pv1.attending_prefix,
                            codification=seg_pv1.attending_type_code,
                        )
                    for seg_tq1 in seg_orc.sample_priority:
                        for seg_obr in seg_tq1.orders:
                            if seg_orc.order_control == "CA":
                                action = "del"
                            else:
                                action = "new"
                            if seg_obr.collector_last_name:
                                sample.physicians.get_or_create(
                                    type="sampler",
                                    code=seg_obr.collector_id,
                                    last_name=seg_obr.collector_last_name,
                                    firt_name=seg_obr.collector_first_name,
                                    title=seg_obr.collector_prefix,
                                    codification=seg_obr.collector_id_coding,
                                )
                            sample.orders.get_or_create(
                                code=seg_obr.test_id,
                                priority=seg_tq1.priority,
                                action=action,
                                order_id=seg_obr.place_order_number,
                            )
                            # specimens infos
                            if seg_obr.specimens:
                                for seg_spm in seg_obr.specimens:
                                    sample.specimens.get_or_create(
                                        code=seg_spm.specimen_type,
                                        label=seg_spm.specimen_label,
                                        barcode=seg_spm.specimen_id,
                                        collection_date=seg_spm.collection_date,
                                        received_date=seg_spm.received_date,
                                        number=seg_spm.number_of_tubes,
                                    )
                            # results
                            if seg_obr.results:
                                for seg_obx in seg_obr.results:
                                    if seg_obx.value_type == "RP":
                                        sample.results.add(
                                            code=seg_obx.observation_identifier,
                                            label=seg_obx.observation_text,
                                            alt_code=seg_obx.alternate_identifier,
                                            alt_label=seg_obx.alternate_text,
                                            value=seg_obx.value.line1,
                                            date=seg_obx.date,
                                        )
                                    elif seg_obx.value_type == "ED":
                                        data = base64.b64decode(seg_obx.value.line5)
                                        values = [
                                            seg_obx.value.line1,
                                            seg_obx.value.line2,
                                            seg_obx.value.line3,
                                            seg_obx.value.line4,
                                            data,
                                        ]
                                        sample.results.add(
                                            code=seg_obx.observation_identifier,
                                            label=seg_obx.observation_text,
                                            alt_code=seg_obx.alternate_identifier,
                                            alt_label=seg_obx.alternate_text,
                                            value=values,
                                            date=seg_obx.date,
                                        )
                                    else:
                                        unit = ""
                                        if seg_obx.units:
                                            unit = seg_obx.units
                                        sample.results.add(
                                            code=seg_obx.observation_identifier,
                                            label=seg_obx.observation_text,
                                            value=seg_obx.value.line1,
                                            unit=unit,
                                            date=seg_obx.date,
                                        )

        return worklist
