# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime
import csv
import logging

from six import StringIO

from driverng import MessageV2, Worklist
from driverng.splitted import Multiple, Splitted, String, Date


logger = logging.getLogger("driverng.drivers.lifecycle_perkinelmer")


class LifeCyclePerkinElmerResultsBase(Splitted):

    SEPARATOR = ";"


class LifeCyclePerkinElmerOrder(LifeCyclePerkinElmerResultsBase):
    FIELDS = (
        String("patient_id"),
        String("mother_last_name"),
        String("mother_first_name"),
        String("mother_address1"),
        String("mother_address2"),
        String("post_code"),
        String("city"),
        String("ethnicity"),
        String("fertilization_date"),
        Date("birthdate", fmt_date="%Y%m%d"),
        String("mother_weight"),
        String("diabetes_did_status"),
        String("smoking_status"),
        String("ante_t21"),
        String("number_fetuses"),
        String("date_last_mentrual"),
        String("hospital"),
        String("doctor"),
        String("doctor2"),
        String("gestional_age"),
        String("sample_id"),
        Date("collector_date", fmt_date="%Y%m%d"),
        String("scan_date"),
        String("nt_mesureament"),
        String("nt_mesureament_second_fetus"),
        String("crl_mesureament"),
        String("crl_mesureament_second_fetus"),
        # Nom de jeune fille
        String("patient_alternative_name"),
        String("monozygous"),
        Date("reception_date", fmt_date="%Y%m%d"),
        String("donor_birthdate"),
        String("age_extraction"),
        String("age_accuracy"),
        String("extraction_date"),
        String("transfert_date"),
        String("assistance_method"),
        String("pre_eclampsia"),
        String("conception_method"),
        String("mother_patient_preeclampsia"),
        String("prev_preg_preeclampsia"),
        String("pregnancies"),
        String("previous_preg_delivery_date"),
        String("prev_preg_gest_delivery"),
        String("prev_preg_baby_weight"),
        String("past_pre_term_delivery"),
        String("chronic_hypertension"),
        String("diabetes_type_2"),
        String("insulin_treatment"),
        String("past_gestational_diabetes"),
        String("anti_phospholipid"),
        String("Height"),
        String("Weight_kg_1"),
        String("Weight_kg_2"),
        String("time_measured"),
        String("test_value"),
        String("right_uterine_artery"),
        String("left_uterine_artery"),
        String("right_diast_1"),
        String("right_diast_2"),
        String("left_diast_1"),
        String("left_diast_2"),
        String("right_syst_1"),
        String("right_syst_2"),
        String("left_syst_1"),
        String("left_syst_2"),
    )


class LifeCyclePerkinElmerResult(object):
    FIELDS = [
        String("patient_id"),
        String("mother_last_name"),
        String("mother_first_name"),
        String("mother_birthdate"),
        String("date_last_menstrual"),
        String("number_fetuses"),
        String("previous_downs_syndrome"),
        String("fertilization_date"),
        String("diabetes_status"),
        String("mother_smoking_status"),
        String("gestational_age"),
        String("sample_id"),
        String("date_sample_collected"),
        Multiple("results", (String("resu"),), separator=";"),
    ]

    def parse(self, line):
        fields = self.FIELDS
        for i, field in enumerate(fields):
            if isinstance(field, Multiple):
                setattr(self, field.name, line[i:])
                break
            else:
                setattr(self, field.name, line[i])


class LifeCyclePerkinElmerMultiple(list):
    _MESSAGE_TYPE_MAP = {
        MessageV2.ORDER: LifeCyclePerkinElmerOrder,
        MessageV2.RESULT: LifeCyclePerkinElmerResult,
    }

    def add(self, type_message=None, **kwargs):
        obj = self._MESSAGE_TYPE_MAP[type_message](**kwargs)
        self.append(obj)
        return obj

    def parse(self, data):
        buf = StringIO(data)
        reader = csv.reader(buf, delimiter=str(";"))
        for line in reader:
            if not len(line):
                continue
            if len(line) == 65:
                type_message = MessageV2.ORDER
            else:
                type_message = MessageV2.RESULT
            obj = self.add(type_message=type_message)
            obj.parse(line)

    def to_string(self):
        return "\n".join([e.to_string() for e in self])


class Driver(MessageV2):
    SPEC = {"code": "lifecycle_perkinelmer", "label": "LifeCycle PerkinElmer"}

    OPTIONS = [
        MessageV2.Option(
            "lifecycle_fumeur_code", label="Code demande FUMEUR", default="FUMEUR"
        ),
        MessageV2.Option(
            "lifecycle_poids_code", label="Code demande POIDS", default="POIDS"
        ),
        MessageV2.Option(
            "lifecycle_diabete_code", label="Code demande DIABETE", default="DIABETE"
        ),
        MessageV2.Option(
            "lifecycle_clarte_nucale_code",
            label="Code demande Clarté Nucale",
            default="CN",
        ),
        MessageV2.Option(
            "lifecycle_date_grossesse_code",
            label="Code demande Date de Grossesse",
            default="DDG",
        ),
        MessageV2.Option(
            "lifecycle_date_echographie_code",
            label="Code demande Date d'échographie",
            default="DE",
        ),
        MessageV2.Option(
            "lifecycle_lcc_code",
            label="Code demande longeur cranio caudale (LCC)",
            default="LCC",
        ),
        MessageV2.Option(
            "lifecycle_code_echographiste",
            label="Code demande du code échographiste",
            default="ECHOGRAPHISTE",
        ),
        MessageV2.Option(
            "lifecycle_code_ethnie", label="Code demande de l'ethnie", default="ETHNIE"
        ),
        MessageV2.Option(
            "lifecycle_code_ante_t21",
            label="Code demande de l'antériorité T21",
            default="ANTE_T21",
        ),
        MessageV2.Option(
            "lifecycle_ddn_donneuse",
            label="Code demande date de naissance de la donneuse",
            default="TR21D_DDN",
        ),
        MessageV2.Option(
            "lifecycle_age_donneuse", label="Age de la donneuse", default="TR21T_AGE"
        ),
        MessageV2.Option(
            "lifecycle_date_prelevement_ovocyte",
            label="Code demande date de prélèvement ovocyte",
            default="TR21P_PREL",
        ),
        MessageV2.Option(
            "lifecycle_date_transfert_ovocyte",
            label="Code demande date de transfert ovocyte",
            default="TR21P_TRANSF",
        ),
        MessageV2.Option(
            "lifecycle_methode_fiv", label="Méthode FIV", default="TR21M_AMPT"
        ),
    ]

    def __init__(self, **kwargs):
        self.fumeur_code = kwargs.pop("lifecycle_fumeur_code", "FUMEUR")
        self.poids_code = kwargs.pop("lifecycle_poids_code", "POIDS")
        self.diabete_code = kwargs.pop("lifecycle_diabete_code", "DIABETE")
        self.clarte_nucale_code = kwargs.pop("lifecycle_clarte_nucale_code", "CN")
        self.date_grossesse_code = kwargs.pop("lifecycle_date_grossesse_code", "DDG")
        self.date_echographie_code = kwargs.pop("lifecycle_date_echographie_code", "DE")
        self.lcc_code = kwargs.pop("lifecycle_lcc_code", "LCC")
        self.echographiste = kwargs.pop("lifecycle_code_echographiste", "ECHOGRAPHISTE")
        self.ethnie = kwargs.pop("lifecycle_code_ethnie", "ETHNIE")
        self.ante_t21 = kwargs.pop("lifecycle_code_ante_t21", "ANTE_T21")
        self.ddn_donneuse = kwargs.pop("lifecycle_ddn_donneuse", "TR21D_DDN")
        self.age_donneuse = kwargs.pop("age_donneuse", "TR21T_AGE")
        self.date_prelevement_ovocyte = kwargs.pop(
            "date_prelevement_ovocyte", "TR21P_PREL"
        )
        self.date_transfert_ovocyte = kwargs.pop(
            "date_transfert_ovocyte", "TR21P_TRANSF"
        )
        self.methode_fiv = kwargs.pop("methode_fiv", "TR21M_AMPT")
        super(Driver, self).__init__(**kwargs)

    def create(self):
        return LifeCyclePerkinElmerMultiple()

    def to_string(self, message):
        return message.to_string()

    def parse(self, data):
        obj = LifeCyclePerkinElmerMultiple()
        obj.parse(data)
        return obj

    def encode_worklist(self, worklist, context):
        values_dict = {
            "NON": "0",
            "NO": "0",
            "NONFUMEUSE": "0",
            "NONFUMEUR": "0",
            "OUI": "1",
            "YES": "1",
            "FUMEUSE": "1",
            "FUMEUR": "1",
        }
        ethnicity_dict = {
            "AFRIQ": "51",
            "ASIAT": "52",
            "CAUCA": "53",
            "ORIENT": "54",
            "AUTRE": "60",
            "INC": "0",
        }
        message = self.create()
        samples = worklist.get_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        )
        if len(samples) == 0:
            return None
        elif len(samples) > 1:
            raise RuntimeError("Multi-sample not implemented")
        else:
            sample = samples[0][0]
            physician = [p for p in sample.physicians if p.type == "prescriber"]
            smoking_status = ""
            diabete_status = ""
            nt_mesureament = ""
            crl_mesureament = ""
            fertilization_date = ""
            scan_date = ""
            mother_weight = ""
            echographiste = ""
            ethnicity = ""
            ante_t21 = ""
            donor_birthdate = ""
            age_extraction = ""
            extraction_date = ""
            transfert_date = ""
            assistance_method = ""
            for result in sample.results:
                if self.fumeur_code and result.code == self.fumeur_code:
                    smoking_status = values_dict.get(result.value)
                if self.poids_code and result.code == self.poids_code:
                    mother_weight = result.value
                elif sample.weight:
                    mother_weight = sample.weight
                if self.diabete_code and result.code == self.diabete_code:
                    diabete_status = values_dict.get(result.value)
                # Clarté nucale
                if self.clarte_nucale_code and result.code == self.clarte_nucale_code:
                    nt_mesureament = result.value
                # Longeur cranio caudale (LCC)
                if self.lcc_code and result.code == self.lcc_code:
                    crl_mesureament = result.value
                # Date de grossesse
                if (
                    self.date_grossesse_code
                    and result.code == self.date_grossesse_code
                    and len(result.value) >= 8
                ):
                    fertilization_date = result.value[:8]
                # Date d'échographie
                if (
                    self.date_echographie_code
                    and result.code == self.date_echographie_code
                    and len(result.value) >= 8
                ):
                    scan_date = result.value[:8]
                if self.echographiste and result.code == self.echographiste:
                    echographiste = result.value
                if self.ethnie and result.code == self.ethnie:
                    ethnicity = ethnicity_dict.get(result.value)
                if self.ante_t21 and result.code == self.ante_t21:
                    ante_t21 = values_dict.get(result.value)
                # Date de naissance de la donneuse
                if (
                    self.ddn_donneuse
                    and result.code == self.ddn_donneuse
                    and len(result.value) >= 8
                ):
                    donor_birthdate = result.value[:8]
                # Age de la donneuse
                if self.age_donneuse and result.code == self.age_donneuse:
                    age_extraction = result.value
                # Date du prélèvement des ovocytes
                if (
                    self.date_prelevement_ovocyte
                    and result.code == self.date_prelevement_ovocyte
                    and len(result.value) >= 8
                ):
                    extraction_date = result.value[:8]
                # Date du transfert ovocyte
                if (
                    self.date_transfert_ovocyte
                    and result.code == self.date_transfert_ovocyte
                    and len(result.value) >= 8
                ):
                    transfert_date = result.value[:8]
                # Méthode de la FIV
                if self.methode_fiv and result.code == self.methode_fiv:
                    assistance_method = result.value
            message.add(
                type_message=MessageV2.ORDER,
                patient_id=sample.patient_id,
                mother_last_name=sample.last_name,
                mother_first_name=sample.first_name,
                mother_address1=sample.address.strip(),
                city=sample.city,
                post_code=sample.postal_code,
                ethnicity=ethnicity,
                ante_t21=ante_t21,
                birthdate=sample.birthdate,
                patient_alternative_name=sample.maiden_name,
                sample_id=sample.sample_id,
                collector_date=sample.collection_date,
                reception_date=sample.request_date,
                # Code du prescripteur
                hospital=physician[0].code if physician else "",
                doctor2=echographiste,
                diabetes_did_status=diabete_status,
                smoking_status=smoking_status,
                nt_mesureament=nt_mesureament,
                crl_mesureament=crl_mesureament,
                fertilization_date=fertilization_date,
                scan_date=scan_date,
                mother_weight=mother_weight,
                donor_birthdate=donor_birthdate,
                age_extraction=age_extraction,
                extraction_date=extraction_date,
                transfert_date=transfert_date,
                assistance_method=assistance_method,
            )
            return message

    def decode_message(self, message):
        worklist = Worklist(context="result")
        for m in message:
            sample = worklist.get_or_create(sample_id=m.sample_id, type="patient")
            if m.date_last_menstrual:
                date = datetime.datetime.strptime(m.date_last_menstrual, "%Y%m%d")
                value = date.strftime("%d/%m/%Y")
                sample.results.add(code="DATE_DERNIERE_REGLE", value=value)
            if m.number_fetuses:
                sample.results.add(code="NOMBRE_GEMELLAIRE", value=m.number_fetuses)
            if m.date_sample_collected:
                date = datetime.datetime.strptime(m.date_sample_collected, "%Y%m%d")
                value = date.strftime("%d/%m/%Y")
                sample.results.add(code="DATE_PRELEVEMENT", value=value)
            if m.previous_downs_syndrome:
                sample.results.add(code="ANTE_T21", value=m.previous_downs_syndrome)
            if m.fertilization_date:
                date = datetime.datetime.strptime(m.fertilization_date, "%Y%m%d")
                value = date.strftime("%d/%m/%Y")
                sample.results.add(code="DATE_DEBUT_GROSSESSE", value=value)
            if m.diabetes_status:
                sample.results.add(code="DIABETES_DID", value=m.diabetes_status)
            if m.mother_smoking_status:
                sample.results.add(code="FUMEUR", value=m.mother_smoking_status)
            if m.gestational_age:
                sample.results.add(code="AGE_GESTATIONEL", value=m.gestational_age)
            for index, result in enumerate(m.results, 13):
                if index in (13, 35, 38, 41, 42) and result and len(result) == 8:
                    date = datetime.datetime.strptime(result, "%Y%m%d")
                    value = date.strftime("%d/%m/%Y")
                else:
                    value = result.replace(",", ".")
                if result:
                    sample.results.add(
                        code="FIELD_RESULT_{}".format(index), value=value.strip()
                    )
        return worklist
